/*
 *				**** Get Info ****
 *
 * This is a sample program to help students learn how to use
 * various packages that are available on Bill Pringle's website
 */

/* standard includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* include file for Bill Pringle's packages */
#include "prodlist.h"
#include "read_line.h"
#include "linklist.h"

/* include file for this program */
#include "getinfo.h"

#include "stuff.h"

#define MAXNAME 50

int debugsw = 9;

/* prototypes for local routines */
void enter_stuff(ProdPtrList *ent);	/* enter entries interactively */
void display_entries(ProdPtrList *ent); /* display all entries */
int get_option(char *buf);		/* promput user for option */
void index_name(ProdPtrList *list, LinkList *ndx);	/* build name index */
void list_names(LinkList *list);
ProdPtrList *find_name(ProdPtrList *list, LinkList *ndx);	/* find using name index */
void save_entries(ProdPtrList *list);
void load_entries(ProdPtrList *list);
void append_entries(ProdPtrList *list);
void insert_name(LinkList *ndx, struct stuff *ent);
void build_full_index(LinkList *list, ProdAPtrList *ndx);
void list_full_index(ProdAPtrList *list);
void build_part_index(LinkList *list, Link *ndx[]);
void list_part_index(Link *ndx[]);

int main(int argc, char *argv[])
{
	char buf[BUFSIZ+1];
   ProdPtrList *stuff;
   LinkList *name_ndx=NULL;	/* index for names */
   struct stuff *ent;
   int i, opt, running=1;
   ProdAPtrList *fullndx;
   struct stuff *partndx[28];	/* a-z, plus values before & after */

   buf[BUFSIZ] = '\0';	/* prevent buffer overflow */

   printf("GetInfo Program, version %s\n", VERSION);

   /* create list for the collection */
   stuff = create_ptr_prodlist(100);	/* starting size */
   name_ndx = create_link_list();	/* used to index names */
   fullndx = create_aptr_prodlist(100);

   while (running)
   {
   	opt = get_option(buf);
      if (*buf == 'q') opt = OPT_DONE;	/* special shortcut */
      switch(opt)
      {
      case OPT_ENTER:
      	enter_stuff(stuff);
      	break;
      case OPT_DISPLAY:
      	display_entries(stuff);
         break;
      case OPT_NDX_NAME:
			index_name(stuff, name_ndx);	/* build name index */
         build_full_index(name_ndx, fullndx);
		   if (debugsw > 5)
         {
         	list_full_index(fullndx);
	         printf("Press ENTER to continue: ");
	         gets(buf);
         }
         build_part_index(name_ndx, partndx);
         if (debugsw > 5)
         {
         	list_part_index(partndx);
            printf("Press ENTER to continue: ");
            gets(buf);
         }
      	break;
      case OPT_SRCH_NAME:
      	break;
      case OPT_LIST_NAME:
      	list_names(name_ndx);
      	break;
      case OPT_SAVE:
      	save_entries(stuff);
      	break;
      case OPT_LOAD:
      	load_entries(stuff);
         break;
      case OPT_APPEND:
      	append_entries(stuff);
      	break;
      case OPT_DONE:
      	printf("\nAre you sure: (y or n): ");
         gets(buf);
			if (*buf == 'y') running=0;
         break;
      default:
      	printf("Invalid option (%d)\n", opt);
      }
   }

   return 0;
}

void display_entries(ProdPtrList *stuff)
{
   struct stuff *ent;
	int i, j, n1, n2;
   char buf[BUFSIZ+1];

   buf[BUFSIZ] = '\0'; /* prevent buffer overflow */

   /* get range to display */
   printf("Enter first entry to display (0-%d): ", stuff->nitems-1);
   gets(buf);
   n1 = atoi(buf);
   if (n1 < 0 || *buf == '\0') return;	/* bad value - return */
   printf("Enter last entry to display (%d-%d): ", n1, stuff->nitems-1);
   gets(buf);
   n2 = atoi(buf);

   /* display requested entries */
   for (i=n1,j=0; i <= n2; i++)
   {
   	ent = (struct stuff *)stuff->item[i];	/* get item */
      display_stuff(ent);
      if ((++j % 20) == 0)
      {
      	printf("Press ENTER to continue: ");
         gets(buf);
         if (*buf == 'q') break;
      }
	}
}

void enter_stuff(ProdPtrList *stuff)
{
	char buf[BUFSIZ+1];
   struct stuff *ent;

   buf[BUFSIZ] = '\0';	/* prevent buffer overflow */

   /* get list of entries from user */
   while (1)
   {
    	printf("Enter ID, or ENTER to quit: ");
      read_line(buf, stdin);
      if (*buf == '\0') break;	/* done entering info */
      /* create entry and fill in fields */
      ent = create_stuff();
      ent->id = atoi(buf);
      printf("Enter last name: ");
      read_line(buf, stdin);
      ent->lname = strdup(buf);	/* last name */
      printf("Enter first name: ");
      read_line(buf, stdin);
      ent->fname = strdup(buf);
      printf("Enter phone number: ");
      read_line(buf, stdin);
      ent->phone = strdup(buf);

      /* now add entry to list */
      add_ptr_prodlist(stuff, ent);
   }
}

int get_option(char *buf)
{
	int opt;

   printf("\nOptions are:\n");
   printf("%2d. Enter entries\n", OPT_ENTER);
   printf("%2d. Display entries\n", OPT_DISPLAY);
   printf("%2d. Build Name index\n", OPT_NDX_NAME);
   printf("%2d. Search by Name\n", OPT_SRCH_NAME);
   printf("%2d. Save entries\n", OPT_SAVE);
   printf("%2d. Load from file\n", OPT_LOAD);
   printf("%2d. Append from file\n", OPT_APPEND);
	printf("%2d. Quit\n", OPT_DONE);
   printf("\nEnter option: ");
   gets(buf);
   opt = atoi(buf);

   return opt;
}

void index_name(ProdPtrList *list, LinkList *ndx)
{	/* build name index */
   struct stuff *ent;
   int i;

   /* loop thru ProdPtrList collection and add to linked list */
   for (i=0; i < list->nitems; i++)
   {
   	ent = (struct stuff *)list->item[i];
      if (debugsw > 9)
      {
      	printf("Indexing %d - %s, %s\n", ent->id, ent->lname, ent->fname);
      }
      insert_name(ndx, ent);
   }
}

void insert_name(LinkList *ndx, struct stuff *ent)
{
	struct stuff *ts;
   Link *nlnk, *tlnk;
   
	nlnk = create_link_entry(ent);	/* create link entry */
   for (tlnk=ndx->head; tlnk != NULL; tlnk = tlnk->next)
   {
   	ts = (struct stuff *)get_data_from_link(tlnk);
      if (debugsw > 9)
      {
      	printf("Comparing with %d - %s, %s\n", ts->id, ts->lname, ts->fname);
      }
      if (strcmp(ent->lname, ts->lname) < 0) break; /* less than, insert here */
      if (strcmp(ent->lname, ts->lname) == 0)
      { /* last names match, check first name */
      	if (strcmp(ent->fname, ts->fname) < 0) break;
      }
   }
   /* if first time, just add to front of list */
   if (tlnk == NULL)
   {
   	add_link_to_tail(ndx, nlnk);
   }
   else
   { /* add before or after temp link */
   	if (strcmp(ent->lname, ts->lname) < 0)
      { /* add ent1 in front of ent2 */
      	add_link_before_link(ndx, nlnk, tlnk);
      }
      else if (strcmp(ent->lname, ts->lname) > 0)
      {
      	add_link_after_link(ndx, nlnk, tlnk);
      }
      else
      { /* last names match, check first names */
	      if (strcmp(ent->fname, ts->fname) < 0)
      	{
      		add_link_before_link(ndx, nlnk, tlnk);
      	}
      	else
         {
         	add_link_after_link(ndx, nlnk, tlnk);
         }
      }
   }
}

void list_names(LinkList *list)
{
	Link *lnk;
   struct stuff *ent;
   char buf[BUFSIZ+1];

   if (list == NULL || list_is_empty(list))
   {
   	printf("List is empty!\n");
      return;
   }

   buf[BUFSIZ] = '\0';

	for (lnk=list->head; lnk != NULL; lnk = lnk->next)
   {
   	ent = (struct stuff *)get_data_from_link(lnk);
      display_stuff(ent);
   }

}

ProdPtrList *find_name(ProdPtrList *list, LinkList *ndx)
{	/* find using name index */
	printf("Not implemented yet.\n");
	return NULL;
}

void save_entries(ProdPtrList *list)
{
	char buf[BUFSIZ];
   FILE *fp;
   int i;
   struct stuff *ent;

   printf("\nEnter name of output file: ");
   gets(buf);
   if ((fp = fopen(buf, "w")) == NULL)
   {
   	printf("Can't open output file (%s)\n", buf);
      return;
   }

   save_stuff(fp, list);
   fclose(fp);
}

void load_entries(ProdPtrList *list)
{
	char buf[BUFSIZ];
   FILE *fp;

   printf("Enter name of input file: ");
   gets(buf);
   if (*buf == '\0') return; /* no file - return quietly */
   if ((fp = fopen(buf, "r")) == NULL)
   {
   	printf("Can't open input file (%s)\n", buf);
      return;
   }

   load_stuff(fp, list);
}

void append_entries(ProdPtrList *list)
{
	char buf[BUFSIZ];
   FILE *fp;

   printf("Enter name of input file: ");
   gets(buf);
   if (*buf == '\0') return;	/* no file - return quietly */
   if ((fp = fopen(buf, "r")) == NULL)
   {
   	printf("Can't open input file (%s)\n", buf);
      return;
   }

   append_stuff(fp, list);
}

void build_full_index(LinkList *list, ProdAPtrList *ndx)
{ /* build full index for names */
	Link *lnk;
   struct stuff *ent;
   char buf[BUFSIZ], *key;

   for (lnk = list->head; lnk != NULL; lnk = lnk->next)
   {
   	ent = (struct stuff *)get_data_from_link(lnk);
      key = ent->lname;		/* just use last name as key */
      /*
       * to use a more complex key, we could do:
       *    sprintf(buf, "%s, %s", ent->lname, ent->fname);
       *    key = buf;
       */
      /* see if key already exists */
		if (bsearch_aptr_prodlist(ndx, key) < 0)
      { /* not already there - add to the index */
      	bupdate_aptr_prodlist(ndx, key, lnk);
      }
   }
}

void list_full_index(ProdAPtrList *list)
{
	int i;
   Link *lnk;
   struct stuff *ent;

   printf("\nFull index entries:\n");
   for (i=0; i < list->nitems; i++)
   {
      lnk = (Link *)list->item[i];
   	ent = (struct stuff *)get_data_from_link(lnk);
      printf("Key=%s, rec=", list->key[i]);
      display_stuff(ent);
   }
}

void build_part_index(LinkList *list, Link *ndx[])
{
	int i;
   Link *lnk;
   struct stuff *ent;
   char lastchar='\0', thischar;

   for (i=0; i < 28; i++) ndx[i] = NULL;	/* start with empty table */

   /* loop through list & save location of first entry of any letter */
   for (lnk = list->head; lnk != NULL; lnk = lnk->next)
   {
   	ent = (struct stuff *)get_data_from_link(lnk);
      thischar = get_first_char(ent->lname);	/* first letter of last name */
      if (thischar != lastchar)
      { /* new letter found - add entry */
      	ndx[thischar] = lnk;	/* point to first entry */
         lastchar = thischar;	/* save this entry */
      }
   }
}

int get_first_char(char *lname)
{
	int thischar;

   thischar = lname[0];
	if (islower(thischar)) thischar = toupper(thischar);	/* use upper case */
   if (thischar < 'A') thischar = 0;	/* before alphabet */
   else if (thischar > 'Z') thischar = 27; /* after alphabet */
   else thischar -= ('A'-1); /* index from start of alphabet (base 1) */

   return thischar;
}

void list_part_index(Link *ndx[])
{
	int i;
   struct stuff *ent;
   Link *lnk;

   printf("\nPart Index entries:\n");
   if (ndx[0] != NULL)
   {
   	printf("Entry before 'A': ");
         lnk = (Link *)ndx[0];
         ent = (struct stuff *)get_data_from_link(lnk);
         display_stuff(ent);
   }
   for (i=1; i <= 26; i++)
   { /* loop through index */
   	if (ndx[i] != NULL)
      {
      	printf("Entry '%c': ", i+'A'-1);
         lnk = (Link *)ndx[i];
         ent = (struct stuff *)get_data_from_link(lnk);
         display_stuff(ent);
      }
   }
   if (ndx[27] != NULL)
   {
   	printf("Entry after 'Z': ");
         lnk = (Link *)ndx[27];
         ent = (struct stuff *)get_data_from_link(lnk);
         display_stuff(ent);
   }
}


